// Unit MATRICE

// OBJECT : matrix and vector aritmetic (inversion, addition, multiplication ...)
// first written in Turbo-Pascal (02-05-1991)
// adapted to Delphi 4 (28-05-1999)

// NOTES

// Note 1 : these routines operate on dynamical square matrix and dynamical vectors
// (RCOMat and RCOVec),whose dimension should have been be declared with
// the InitMat(Dim) procedure (error if if not initialize).

// Note 2 : the index (i,j..) are "zero based", this is a constraint of
// Delphi4 dynamical matrix structure.

// Note 3 : the "if true ..." instructions are remainings of the TurboPascal 5 version,
// I should have transform them into "try .. except .." instructions in order to
// manage memory overflow errors. If somebody do that, I can mail me the upgraded version.

// Note 4 : I checked this unit with the small program at the end of that file.
// The time required to invert a 200 x 200 extented matrix is of about 2 seconds,
// (AMD K6-2 overclocked at 400 MHz), and the accuracy compatible with extended
// accuracy (10E-17 .. 10E-18)

// if you find that unit usefull, please let me know :
// remi CORMIER, rco@club-internet.fr

UNIT Matrice;

INTERFACE
// I would'nt worry if you change "RCOMat" to "YourNameMat" ...
TYPE      RCOMat = array of array of extended;
          RCOVec = array of extended;

VAR       ErreurMat : integer;  { code d'erreur, Cf ci aprs :   }

CONST     MatCorrect   = 0;
          MatNonInit   = 1;
          MatDebordTas = 2;
          MatDetNul    = 3;

// Returns (W) = (U)-(V) ; U,V,W : Vectors
PROCEDURE DifVec(U,V:RCOVec;var W:RCOVec);

// Returns the current dimension of matrix and vectors
FUNCTION  DimMatrices:integer;

// Initialize the matrix and vectors dimension with Dim
PROCEDURE InitMat(Dim:integer);

// Returns [B], the invert matrix of [A] ([A]x[B]=[I]), Det is the determinant of [A] (and [B])
PROCEDURE InvMat(A:RCOMat;var B:RCOMat;var Det:extended{MatElem});

// Returns [C] = [A]x[B]
PROCEDURE MatMat(A,B:RCOMat;var C:RCOMat);

// Returns [C] = [A]x[B] ; same as MatMat, except that [A] and [B] can be
// rectangular. [A] have Dim rows and M columns, [B] have M rows and P colums :
// [A]N,M x [B]M,P -> [C]N,P (with N = DimMat)
PROCEDURE MatMatR(A,B:RCOMat;M,P:integer;var C:RCOMat);

// Returns the unit matrix (one on diagonal, zero everywhere else)
PROCEDURE MatUnt(var MU:RCOMat);

// Returns the (W)=[M]x(V) ; [M] RCOMat, (V) and (W) RCOVec
PROCEDURE MatVec(M:RCOMat;V:RCOVec;var W:RCOVec);

// Returns the norm of vector (U)
FUNCTION  Norme(U:RCOVec)    : extended;


// Returns the norm of [A], [A] can be rectangular, with n row and m columns
FUNCTION  NorMat(A:RCOMat;n,m:integer):extended;

// Returns the scalar product of the vector (U) and (V)
FUNCTION  PrdScal(U,V:RCOVec)  : extended;

// Multiply the matrix [A] with x, the result in matrix [B]
PROCEDURE ScalMat(x:extended;A:RCOMat;var B:RCOMat);

// Multiply the vector (U) with x, the result in vector (V)
PROCEDURE ScalVec(x:extended;U:RCOVec;var V:RCOVec);

// Addition of two square matrix [A] and [B], result in [C]
PROCEDURE SomMat(A,B:RCOMat;var C:RCOMat);

// Addition of two vectors, (U) and (V), result in (W)
PROCEDURE SomVec(U,V:RCOVec;var W:RCOVec);

// Solution of the linear system [A]x(X)=(B)
// vector (X) is the solution
// check Det for valid solution (should be <> 0)
PROCEDURE SysLin(A:RCOMat;var X:RCOVec;B:RCOVec;var Det:extended{MatElem});

// Returns the transposed matrix of [A] into [B] (Bij = Aji)
PROCEDURE TransMat(A:RCOMat;var B:RCOMat);

// Returns the text (in french) associated with the error # no
FUNCTION  MsgErreurMat(no:integer):string;

IMPLEMENTATION

VAR   Ini       : boolean;
      DimMat    : integer;

TYPE  RCoVecE   = array of integer;

FUNCTION  MsgErreurMat(no:integer):string;
BEGIN
  MsgErreurMat:='';
  case no of
    MatCorrect   : Result:='succs';
    MatNonInit   : Result:='unit non initialise';
    MatDebordTas : Result:='pas assez de mmoire';
    MatDetNul    : Result:='dterminant nul';
    end;
END;

FUNCTION  NorMat(A:RCOMat;n,m:integer):extended;
VAR Nor : extended;
    i,j : integer;
BEGIN
  Nor:=0;
  for i:=0 to pred(n) do
    for j:=0 to pred(m) do
      Nor:=Nor*sqr(A[i,j]);
  Nor:=sqrt(Nor/n/m);
  Result:=Nor;
END;

FUNCTION  DimMatrices:integer;
BEGIN
  DimMatrices:=DimMat;
END;

PROCEDURE Init;
begin
  if not Ini then
    begin
      ErreurMat:=MatDebordTas;
      write('Unit MATRICES non initialise');// I think that write generate the write
      halt(1);                                // instruction would generate an exception
    end;
  ErreurMat:=0;
end;

PROCEDURE InitMat(Dim:integer);
begin
  DimMat:=Dim;
  Ini:=true;
  Init;
end;

PROCEDURE MatMat(A,B:RCOMat;var C:RCOMat);
var i, j , k : integer;
    Cij      : extended;

begin
  Init;
  if true
    then
      begin
        for i:=0 to pred(DimMat)  do
          for j:=0 to pred(DimMat) do
            begin
              Cij:=0;
              for k:=0 to pred(DimMat) do
                Cij:=Cij+A[i,k]*B[k,j];
              C[i,j]:=Cij;
            end;
        ErreurMat:=0;
      end
    else ErreurMat:=MatDebordTas;
end;

PROCEDURE SomMat(A,B:RCOMat;var C:RCOMat);
var i,j : integer;

BEGIN
  if true
    then
      begin
        for i:=0 to pred(DimMat) do
          for j:=0 to pred(DimMat) do
            C[i,j]:=A[i,j]+B[i,j];
        ErreurMat:=0;
      end
    else ErreurMat:=MatDebordTas;
END;

PROCEDURE MatMatR(A,B:RCOMat;M,P:integer;var C:RCOMat);
var i,j,k       : integer;
    Cij         : extended;
BEGIN
  if true
    then
      begin
        for i:=0 to pred(DimMat) do
          for j:=0 to pred(P) do
            begin
              Cij:=0;
              for k:=0 to pred(M) do
                Cij:=Cij+A[i,k]*B[k,j];
              C[i,j]:=Cij;
            end;
        ErreurMat:=0;
      end
    else ErreurMat:=MatDebordTas;
END;

PROCEDURE TransMat(A:RCOMat;var B:RCOMat);
var i, j     : integer;
begin
  Init;
  if true
    then
      begin
        for i:=0 to pred(DimMat)  do
          for j:=0 to pred(DimMat) do
            B[i,j]:=A[j,i];
        ErreurMat:=0;
      end
    else ErreurMat:=MatDebordTas;
end;

PROCEDURE MatUnt(var MU:RCOMat);
var i,j     : integer;
begin
  Init;
  for i:=0 to pred(DimMat) do
    for j:=0 to pred(DimMat) do
      if i=j then MU[i,j]:=1
             else MU[i,j]:=0;
end;

PROCEDURE MatVec(M:RCOMat;V:RCOVec;var W:RCOVec);
var i, j     : integer;
    Wi       : extended;
begin
  Init;
  if true
    then
      begin
        for i:=0 to pred(DimMat) do
          begin
            Wi:=0;
            for j:=0 to pred(DimMat) do
              Wi:=Wi+M[i,j]*V[j];
            W[i]:=Wi;
          end;
        ErreurMat:=0;
      end
    else ErreurMat:=MatDebordTas;
end;

FUNCTION  PrdScal(U,V:RCOVec)  : extended;
var i : integer;
    s : extended;
begin
  Init;
  s:=0;
  for i:=0 to pred(DimMat) do
    s:=s+U[i]*V[i];
  Result:=s;
end;

PROCEDURE ScalVec(x:extended;U:RCOVec;var V:RCOVec);
var i        : integer;
begin
  Init;
  for i:=0 to pred(DimMat) do
    V[i]:=x*U[i];
end;

PROCEDURE ScalMat(x:extended;A:RCOMat;var B:RCOMat);
var i,j : integer;
BEGIN
  for i:=0 to pred(DimMat) do
    for j:=0 to pred(DimMat) do
      B[i,j]:=x*A[i,j];
END;

PROCEDURE AddVec (Signe:integer;U,V:RCOVec;var W:RCOVec);
var i        : integer;
begin
  Init;
  for i:=0 to pred(DimMat) do
    W[i]:=U[i]+Signe*V[i];
end;

PROCEDURE SomVec(U,V:RCOVec;var W:RCOVec);
begin
  AddVec(1,U,V,W);
end;

PROCEDURE DifVec(U,V:RCOVec;var W:RCOVec);
begin
  AddVec(-1,U,V,W);
end;

FUNCTION  Norme(U:RCOVec)    : extended;
var i : integer;
    n : extended;
begin
  Init;
  n:=0;
  for i:=0 to pred(DimMat) do
    n:=n+sqr(U[i]);
  Result:=sqrt(n);
end;

PROCEDURE TrianguleO(A:RCOMat;var AT:RCOMat;var Jx,Iy:RCOVecE;var Det:extended);
var i, k     : integer;
    im , jm  : integer;
    ATkk     : extended;
    C        : extended;
    Max      : extended;
    SignDet  : integer;

  PROCEDURE InitJxIy;
  var i:integer;
  begin
    for i:=0 to pred(DimMat) do
    begin
      Jx[i]:=i;
      Iy[i]:=i;
    end;
  end;

  PROCEDURE CopiAAT;
  var i,j:integer;
  begin
    for i:=0 to pred(DimMat) do
    for j:=0 to pred(DimMat) do
      AT[i,j]:=A[i,j];
  end;

  PROCEDURE Reduit;
  var i,j:integer;
  begin
    ATkk:=AT[k,k];
      for i:=k+1 to pred(DimMat) do
        begin
          C   :=AT[i,k]/ATkk;
          for j:=k+1 to pred(DimMat) do
            AT[i,j]:=AT[i,j]-C*AT[k,j];
          for j:=0 to k do
            if j<>k
              then AT[i,j]:=AT[i,j]-C*AT[k,j]
              else AT[i,j]:=-C;
        end;
  end;

  PROCEDURE ChercheMax(var Max:extended;var im,jm:integer);
  var i,j : integer;
      M   : extended;
  begin
    Max:=abs(AT[k,k]);
    im:=k;jm:=k;
    for i:=k to pred(DimMat) do
      for j:=k to pred(DimMat) do
        begin
          M:=abs(AT[i,j]);
          if M>Max then
            begin
              Max:=M;
              im:=i;jm:=j;
            end;
        end;
  end;

  PROCEDURE PermuteCol;
  var Ind : integer;
      i   : integer;
      x   : extended;
  begin
    for i:=0 to pred(DimMat) do
      begin
        x:=AT[i,k];
        AT[i,k]:=AT[i,jm];
        AT[i,jm]:=x;
      end;
    Ind   :=Jx[k];
    Jx[k] :=Jx[jm];
    Jx[jm]:=Ind;
    SignDet:=-SignDet;
  end;

  PROCEDURE PermuteLig;
  var Ind : integer;
      j   : integer;
      x   : extended;
  begin
    for j:=0 to pred(DimMat) do
      begin
        x:=AT[k,j];
        AT[k,j]:=AT[im,j];
        AT[im,j]:=x;
      end;
    Ind:=Iy[k];
    Iy[k]:=Iy[im];
    Iy[im]:=ind;
    SignDet:=-SignDet;
  end;

begin
  InitJxIy;
  CopiAAT;
  k:=0;
  SignDet:=1;
  while (k<=pred(DimMat)-1) and (ErreurMat=0) do
    begin
      ChercheMax(Max,im,jm);
      if Max<>0
        then
          begin
            if k<>jm then PermuteCol;
            if k<>im then PermuteLig;
            Reduit;
          end
        else ErreurMat:=MatDetNul;
      k:=k+1;
    end;
  Det:=SignDet;
  for i:=0 to pred(DimMat) do
    Det:=Det*AT[i,i];
  if Det=0 then ErreurMat:=MatDetNul;
end;

PROCEDURE ResouTrO(A:RCOMat;var X,Y:RCOVec;Jx,Iy:RCOVecE);
var i, j     : integer;
    PVX      : RCOVec;
    Xi       : extended;

begin
  if true
    then
      begin
        SetLength(PVX,DimMat);
        for i:=pred(DimMat) downto 0 do
          begin
            Xi:=Y[Iy[i]];
            for j:=0 to i-1 do
              Xi:=Xi+A[i,j]*Y[Iy[j]];
            for j:=i+1 to pred(DimMat) do
              Xi:=Xi-A[i,j]*PVX[j];
            PVX[i]:=Xi/A[i,i];
          end;
        for i:=0 to pred(DimMat) do
          X[Jx[i]]:=PVX[i];
        Setlength(PVX,0);
        ErreurMat:=0;
      end
    else ErreurMat:=MatDebordTas;
end;

PROCEDURE SysLin(A:RCOMat;var X:RCOVec;B:RCOVec;var Det:extended);
var AT       : RCOMat;
    Jx,Iy    : RCOVecE;

  PROCEDURE Alloc;
  begin
    SetLength(AT,DimMat,DimMat);
    SetLength(Jx,DimMat);
    SetLength(Iy,DimMat);
  end;

  PROCEDURE DesAlloc;
  begin
    SetLength(Iy,0);
    SetLength(Jx,0);
    SetLength(At,0,0);
  end;

begin
  Init;
  if true
    then
      begin
        Alloc;
        TrianguleO(A,AT,Jx,Iy,Det);
        if Det<>0
          then ResouTrO(AT,X,B,Jx,Iy);
        DesAlloc;
      end
    else ErreurMat:=MatDebordTas;
end;

PROCEDURE InvMat(A:RCOMat;var B:RCOMat;var Det:extended);
var j        : integer;
    AT       : RCOMAt;
    PVX      : RCOVec;
    Jx,Iy    : RCOVecE;

  PROCEDURE Alloc;
  begin
    Setlength(AT,DimMat,DimMat);
    SetLength(PVX,DimMat);
    SetLength(Jx,DimMat);
    SetLength(Iy,DimMat);
  end;

  PROCEDURE DesAlloc;
  begin
    SetLength(Iy,0);
    SetLength(Jx,0);
    SetLength(PVX,0);
    SetLength(AT,0,0);
  end;

  PROCEDURE initVX;
  var i : integer;
  begin
    for i:=0 to pred(DimMat) do
      PVX[i]:=0;
    PVX[j]:=1;
  end;

  PROCEDURE TransfertCol;
  var i : integer;
  begin
    for i:=0 to pred(DimMat) do
      B[i,j]:=PVX[i];
  end;

begin
  Init;
  if true
    then
      begin
        Alloc;
        TrianguleO(A,AT,Jx,Iy,Det);
        if ErreurMat=0
          then
            begin
              for j:=0 to pred(DimMat) do
              begin
                InitVX;
                ResouTrO(AT,PVX,PVX,Jx,Iy);
                TransfertCol;
              end;
            end;
        DesAlloc;
      end
    else ErreurMat:=MatDebordTas;
end;

initialization
  Ini:=false;
  DimMat:=0;
END.


// Test program :

procedure TForm1.FormCreate(Sender: TObject);
Const Dim = 200;
var   A,B,C : RCOMat;
      V,W,U : RCOVec;
      i,j   : integer;
      Det   : extended;
begin
  InitMat(Dim);
  SetLength(A,Dim,Dim);
  SetLength(B,Dim,Dim);
  SetLength(C,Dim,Dim);
  Randomize;
  for i:=0 to pred(Dim) do
    begin
      V[i]:=random*10;
      for j:=0 to pred(Dim) do
      A[i,j]:=random*10;
    end;
  InvMat(A,B,Det);
  MatMat(A,B,C);
end;

// [C] should be the unit matrix